﻿/*
   Copyright 2012, Andrew Polar

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace splitandmerge
{
    public class Row
    {
        public List<int> m_sum = new List<int>(0);
        public List<int> m_pos = new List<int>(0);
        public double norm = 0.0;
    }

    public class Cluster
    {
        public List<Row> m_rows = new List<Row>();
        public List<int> m_indexes = new List<int>();
        private List<int> m_left = new List<int>();
        private List<int> m_right = new List<int>();
        private double[] m_leftClassifier = null;
        private double[] m_rightClassifier = null;

        public Cluster()
        {
        }

        public void ClearAll()
        {
            for (int i = 0; i < m_rows.Count; ++i)
            {
                m_rows[i].m_pos.Clear();
                m_rows[i].m_sum.Clear();
            }
            m_rows.Clear();
            m_indexes.Clear();
            m_left.Clear();
            m_right.Clear();
            m_leftClassifier = null;
            m_rightClassifier = null;
        }

        public void AddRow(Row row, int index)
        {
            m_rows.Add(row);
            m_indexes.Add(index);
        }

        public int CountAllElements()
        {
            int nTotal = 0;
            foreach (Row row in m_rows)
            {
                nTotal += row.m_pos.Count;
            }
            return nTotal;
        }

        private double ComputeCosine(int i, int j)
        {
            double product = 0.0;
            int index1 = 0;
            int index2 = 0;
            while (true)
            {
                if (m_rows[i].m_pos[index1] == m_rows[j].m_pos[index2])
                {
                    product += m_rows[i].m_sum[index1] * m_rows[j].m_sum[index2];
                    ++index1;
                    ++index2;
                }
                else if (m_rows[i].m_pos[index1] > m_rows[j].m_pos[index2])
                {
                    ++index2;
                }
                else if (m_rows[i].m_pos[index1] < m_rows[j].m_pos[index2])
                {
                    ++index1;
                }
                if (index1 >= m_rows[i].m_pos.Count) break;
                if (index2 >= m_rows[j].m_pos.Count) break;
            }
            product /= m_rows[i].norm;
            product /= m_rows[j].norm;
            return (float)(product);
        }

        private int FindOpposite(int index)
        {
            double minCosine = 1.0;
            int minIndex = -1;
            for (int i=0; i<m_rows.Count; ++i)
            {
                double cosine = ComputeCosine(index, i);
                if (cosine < minCosine)
                {
                    minCosine = cosine;
                    minIndex = i;
                }
            }
            return minIndex;
        }

        private void Sort2LeftAndRight(int left, int right)
        {
            m_left.Clear();
            m_right.Clear();
            for (int i = 0; i < m_rows.Count; ++i)
            {
                double cosineLeft = ComputeCosine(left, i);
                double cosineRight = ComputeCosine(right, i);
                if (cosineLeft > cosineRight)
                {
                    m_left.Add(i);
                }
                else
                {
                    m_right.Add(i);
                }
            }
        }

        public void ShowSortedVectors()
        {
            for (int i = 0; i < m_left.Count; ++i)
            {
                Console.Write(" {0} ", m_indexes[m_left[i]]);
            }
            Console.WriteLine();
            Console.WriteLine();

            for (int i = 0; i < m_right.Count; ++i)
            {
                Console.Write(" {0} ", m_indexes[m_right[i]]);
            }
            Console.WriteLine();
            Console.WriteLine();
        }

        private void Orthogonalize(ref double[] left, ref double[] right)
        {
            double norm = 0.0;
            int size = left.Length;
            for (int i = 0; i < size; ++i)
            {
                norm += left[i] * left[i];
                norm += right[i] * right[i];
            }
            norm = Math.Sqrt(norm);
            for (int i = 0; i < size; ++i)
            {
                left[i] /= norm;
                right[i] /= norm;
            }
            double[,] tmp = new double[2, 2];
            double[] tmpLeft = new double[left.Length];
            double[] tmpRight = new double[right.Length];

            int counter = 0;
            while (true)
            {
                tmp[0, 0] = 0.0;
                tmp[0, 1] = 0.0;
                tmp[1, 0] = 0.0;
                tmp[1, 1] = 0.0;
                for (int i = 0; i < size; ++i)
                {
                    tmp[0, 0] += left[i] * left[i];
                    tmp[0, 1] += left[i] * right[i];
                    tmp[1, 1] += right[i] * right[i];
                }
                tmp[1, 0] = tmp[1, 0];

                for (int i = 0; i < size; ++i)
                {
                    tmpLeft[i] = left[i] * tmp[0, 0] + right[i] * tmp[0, 1];
                    tmpRight[i] = left[i] * tmp[1, 0] + right[i] * tmp[1, 1];
                }

                for (int i = 0; i < size; ++i)
                {
                    left[i] = 1.5 * left[i] - 0.5 * tmpLeft[i];
                    right[i] = 1.5 * right[i] - 0.5 * tmpRight[i];
                }

                double side = 2.0 * tmp[0, 1];
                double trace = tmp[0, 0] + tmp[1, 1];
                if (side / trace < 0.00001) break;
                if (++counter >= 100) break;
            }
        }

        private void MakeClassifiers(int nSize, int indexLeft, int indexRight)
        {
            m_leftClassifier = new double[nSize];
            m_rightClassifier = new double[nSize];
            for (int i = 0; i < nSize; ++i)
            {
                m_leftClassifier[i] = 0.0;
                m_rightClassifier[i] = 0.0;
            }

            for (int i = 0; i < m_rows[indexLeft].m_pos.Count; ++i)
            {
                m_leftClassifier[m_rows[indexLeft].m_pos[i]] = m_rows[indexLeft].m_sum[i];
            }

            for (int i = 0; i < m_rows[indexRight].m_pos.Count; ++i)
            {
                m_rightClassifier[m_rows[indexRight].m_pos[i]] = m_rows[indexRight].m_sum[i];
            }
            Orthogonalize(ref m_leftClassifier, ref m_rightClassifier);
        }

        private void Sort2LeftAndRightWithClassifiers(int nSize, int leftIndex, int rightIndex)
        {
            MakeClassifiers(nSize, leftIndex, rightIndex);
            m_left.Clear();
            m_right.Clear();
            for (int i = 0; i < m_rows.Count; ++i)
            {
                double leftProduct = 0.0;
                for (int j = 0; j < m_rows[i].m_pos.Count; ++j)
                {
                    leftProduct += m_leftClassifier[m_rows[i].m_pos[j]] * ((double)(m_rows[i].m_sum[j]));
                }

                double rightProduct = 0.0;
                for (int j = 0; j < m_rows[i].m_pos.Count; ++j)
                {
                    rightProduct += m_rightClassifier[m_rows[i].m_pos[j]] * ((double)(m_rows[i].m_sum[j]));
                }

                if (leftProduct > rightProduct)
                {
                    m_left.Add(i);
                }
                else
                {
                    m_right.Add(i);
                }
            }
        }

        private CRow mergeTwoRows(CRow r1, CRow r2)
        {
            CRow merged = new CRow();

            for (int i = 0; i < r1.m_files.Count; ++i)
            {
                merged.m_files.Add(r1.m_files[i]);
            }
            for (int i = 0; i < r2.m_files.Count; ++i)
            {
                merged.m_files.Add(r2.m_files[i]);
            }

            int index1 = 0;
            int index2 = 0;
            while (true)
            {
                if (r1.m_pos[index1] == r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    int sum = r1.m_sum[index1];
                    sum += r2.m_sum[index2];
                    merged.m_sum.Add(sum);
                    ++index1;
                    ++index2;
                }
                else if (r1.m_pos[index1] > r2.m_pos[index2])
                {
                    merged.m_pos.Add(r2.m_pos[index2]);
                    merged.m_sum.Add(r2.m_sum[index2]);
                    ++index2;
                }
                else if (r1.m_pos[index1] < r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    merged.m_sum.Add(r1.m_sum[index1]);
                    ++index1;
                }
                if (index1 >= r1.m_pos.Count) break;
                if (index2 >= r2.m_pos.Count) break;
            }

            for (int i = index1; i < r1.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r1.m_pos[i]);
                merged.m_sum.Add(r1.m_sum[i]);
            }

            for (int i = index2; i < r2.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r2.m_pos[i]);
                merged.m_sum.Add(r2.m_sum[i]);
            }

            merged.norm = 0.0;
            foreach (short s in merged.m_sum)
            {
                merged.norm += s * s;
            }
            merged.norm = Math.Sqrt(merged.norm);

            return merged;
        }

        public CRow PopulateLeft()
        {
            CRow crow1 = new CRow();
            for (int i = 0; i < m_rows[m_left[0]].m_pos.Count; ++i)
            {
                crow1.m_pos.Add(m_rows[m_left[0]].m_pos[i]);
                crow1.m_sum.Add(m_rows[m_left[0]].m_sum[i]);                
            }
            crow1.m_files.Add(m_indexes[m_left[0]]);

            for (int i = 1; i < m_left.Count; ++i)
            {
                CRow crow2 = new CRow();
                for (int j = 0; j < m_rows[m_left[i]].m_pos.Count; ++j)
                {
                    crow2.m_pos.Add(m_rows[m_left[i]].m_pos[j]);
                    crow2.m_sum.Add(m_rows[m_left[i]].m_sum[j]);                    
                }
                crow2.m_files.Add(m_indexes[m_left[i]]);
                crow1 = mergeTwoRows(crow1, crow2);
            }
            return crow1;
        }

        public CRow PopulateRight()
        {
            CRow crow1 = new CRow();
            for (int i = 0; i < m_rows[m_right[0]].m_pos.Count; ++i)
            {
                crow1.m_pos.Add(m_rows[m_right[0]].m_pos[i]);
                crow1.m_sum.Add(m_rows[m_right[0]].m_sum[i]);               
            }
            crow1.m_files.Add(m_indexes[m_right[0]]);

            for (int i = 1; i < m_right.Count; ++i)
            {
                CRow crow2 = new CRow();
                for (int j = 0; j < m_rows[m_right[i]].m_pos.Count; ++j)
                {
                    crow2.m_pos.Add(m_rows[m_right[i]].m_pos[j]);
                    crow2.m_sum.Add(m_rows[m_right[i]].m_sum[j]);                   
                }
                crow2.m_files.Add(m_indexes[m_right[i]]);
                crow1 = mergeTwoRows(crow1, crow2);
            }
            return crow1;
        }

        public void Split(int nVectorSize, bool usePO)
        {
            if (m_indexes.Count == 1)
            {
                m_left.Add(0);
                m_right.Add(0);
                return;
            }

            if (m_indexes.Count == 2)
            {
                m_left.Add(0);
                m_right.Add(1);
                return;
            }

            Random rnd = new Random();
            int random1 = rnd.Next(m_rows.Count);
            int leftPole = FindOpposite(random1);
            int rightPole = FindOpposite(leftPole);

            if (usePO)
            {
                Sort2LeftAndRightWithClassifiers(nVectorSize, leftPole, rightPole);
            }
            else
            {
                Sort2LeftAndRight(leftPole, rightPole);
            }
        }
    }
}
