﻿/*
   Copyright 2012, Andrew Polar

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Collections;

namespace splitandmerge
{
    public class CRow
    {
        public List<int> m_sum = new List<int>(0);
        public List<int> m_pos = new List<int>(0);
        public List<int> m_files = new List<int>(0);
        public double norm = 0.0;
    }

    class HAC
    {
        public List<CRow> m_crows = new List<CRow>(0);
        private List<CRow> m_trainingSet = new List<CRow>(0);

        private string m_documentTermMatrix;

        public HAC(string documentTermMatrix)
        {
            m_documentTermMatrix = documentTermMatrix;
        }

        public void PopulateCRows(List<Cluster> clusterList) 
        {
            foreach (CRow crow in m_crows)
            {
                crow.m_sum.Clear();
                crow.m_files.Clear();
                crow.m_pos.Clear();
            }
            m_crows.Clear();

            foreach (Cluster cluster in clusterList)
            {
                CRow rowLeft = cluster.PopulateLeft();
                CRow rowRight = cluster.PopulateRight();
                m_crows.Add(rowLeft);
                m_crows.Add(rowRight);
            }
        }

        //Debugging print
        public void ShowClusteredFiles()
        {
            Console.WriteLine("Clusters:");
            foreach (CRow crow in m_crows)
            {
                int nCol = 0;
                foreach (int x in crow.m_files)
                {
                    ++nCol;
                    string number = Helper.formatInt(x, 4);
                    Console.Write("{0}", number);
                    if (nCol >= 8)
                    {
                        Console.WriteLine();
                        nCol = 0;
                    }
                }
                Console.WriteLine("\n ------------------------------\n");
            }
        }

        private float getCosine(CRow r1, CRow r2)
        {
            double product = 0.0f;
            int index1 = 0;
            int index2 = 0;
            while (true)
            {
                if (r1.m_pos[index1] == r2.m_pos[index2])
                {
                    product += r1.m_sum[index1] * r2.m_sum[index2];
                    ++index1;
                    ++index2;
                }
                else if (r1.m_pos[index1] > r2.m_pos[index2])
                {
                    ++index2;
                }
                else if (r1.m_pos[index1] < r2.m_pos[index2])
                {
                    ++index1;
                }
                if (index1 >= r1.m_pos.Count) break;
                if (index2 >= r2.m_pos.Count) break;
            }
            product /= r1.norm;
            product /= r2.norm;
            double min = r1.m_files.Count;
            if (min > r2.m_files.Count) min = r2.m_files.Count;
            product /= min;
            return (float)(product);
        }

        private void getMaxElement(ref int nRow, ref int nCol, ref float fMax, float[,] fMatrix, int nSize)
        {
            nRow = 0;
            nCol = 0;
            fMax = fMatrix[nRow, nCol];
            for (int i = 0; i < nSize; ++i)
            {
                for (int j = i + 1; j < nSize; ++j)
                {
                    if (fMax < fMatrix[i, j])
                    {
                        fMax = fMatrix[i, j];
                        nRow = i;
                        nCol = j;
                    }
                }
            }
        }

        private CRow mergeTwoRows(CRow r1, CRow r2)
        {
            CRow merged = new CRow();

            for (int i = 0; i < r1.m_files.Count; ++i)
            {
                merged.m_files.Add(r1.m_files[i]);
            }
            for (int i = 0; i < r2.m_files.Count; ++i)
            {
                merged.m_files.Add(r2.m_files[i]);
            }

            int index1 = 0;
            int index2 = 0;
            while (true)
            {
                if (r1.m_pos[index1] == r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    int sum = r1.m_sum[index1];
                    sum += r2.m_sum[index2];
                    merged.m_sum.Add(sum);
                    ++index1;
                    ++index2;
                }
                else if (r1.m_pos[index1] > r2.m_pos[index2])
                {
                    merged.m_pos.Add(r2.m_pos[index2]);
                    merged.m_sum.Add(r2.m_sum[index2]);
                    ++index2;
                }
                else if (r1.m_pos[index1] < r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    merged.m_sum.Add(r1.m_sum[index1]);
                    ++index1;
                }
                if (index1 >= r1.m_pos.Count) break;
                if (index2 >= r2.m_pos.Count) break;
            }

            for (int i = index1; i < r1.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r1.m_pos[i]);
                merged.m_sum.Add(r1.m_sum[i]);
            }

            for (int i = index2; i < r2.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r2.m_pos[i]);
                merged.m_sum.Add(r2.m_sum[i]);
            }

            merged.norm = 0.0;
            foreach (short s in merged.m_sum)
            {
                merged.norm += s * s;
            }
            merged.norm = Math.Sqrt(merged.norm);

            return merged;
        }

        public void ExecuteHAC(int nCategories)
        {
            int NN = m_crows.Count - nCategories;
            int nMatrixSize = m_crows.Count;
            float[,] fMatrix = new float[nMatrixSize, nMatrixSize];
            for (int i = 0; i < nMatrixSize; ++i)
            {
                for (int j = i + 1; j < nMatrixSize; ++j)
                {
                    fMatrix[i, j] = getCosine(m_crows[i], m_crows[j]);
                }
            }

            for (int loop = 0; loop < NN; ++loop)
            {
                nMatrixSize = m_crows.Count;
                int nrow = 0;
                int ncol = 0;
                float fMax = 0.0f;
                getMaxElement(ref nrow, ref ncol, ref fMax, fMatrix, nMatrixSize);

                CRow r1 = m_crows[nrow];
                CRow r2 = m_crows[ncol];
                CRow merged = mergeTwoRows(r1, r2);
                m_crows.Remove(r1);
                m_crows.Remove(r2);

                //reassign repeated matrix elements
                int curRow = 0;
                int curCol = 0;
                for (int i = 0; i < nMatrixSize; ++i)
                {
                    if (i != nrow && i != ncol)
                    {
                        curCol = curRow + 1;
                        for (int j = i + 1; j < nMatrixSize; ++j)
                        {
                            if (j != nrow && j != ncol)
                            {
                                fMatrix[curRow, curCol] = fMatrix[i, j];
                                ++curCol;
                            }
                        }
                        ++curRow;
                    }
                }
                m_crows.Add(merged);

                //recompute new matrix elements
                curRow = 0;
                for (int i = 0; i < nMatrixSize; ++i)
                {
                    if (i != nrow && i != ncol)
                    {
                        fMatrix[curRow, nMatrixSize - 2] = getCosine(m_crows[curRow], m_crows[nMatrixSize - 2]);
                        ++curRow;
                    }
                }
            }
        }

        public List<int> GetNewLabels()
        {
            int nFiles = 0;
            foreach (CRow crow in m_crows)
            {
                nFiles += crow.m_files.Count;
            }
            List<int> labels = new List<int>(0);
            for (int i = 0; i < nFiles; ++i)
            {
                labels.Add(0);
            }
            
            int cat = 0;
            foreach (CRow crow in m_crows)
            {
                foreach (int x in crow.m_files)
                {
                    labels[x] = cat;
                }
                ++cat;
            }

            return labels;
        }

        public double ComputeF_measure(int[] nExpected, int nCategories)
        {
            List<int> labels = GetNewLabels();
            double f_measure = Helper.Compute_F_Measure(nExpected, labels.ToArray(), true);
            //resort clusters
            int[] lookup = Helper.GetLookup();
            if (lookup != null)
            {
                int max = lookup[0];
                foreach (int x in lookup)
                {
                    if (x > max) max = x;
                }

                for (int i = 0; i < max + 1; ++i)
                {
                    for (int j = 0; j < max + 1; ++j)
                    {
                        if (lookup[j] == i)
                        {
                            m_crows.Add(m_crows[j]);
                        }
                    }
                }

                for (int i = 0; i < max + 1; ++i)
                {
                    m_crows.Remove(m_crows[0]);
                }
            }
            //end

            return f_measure;
        }

        private CRow[] MakeListOfRows(int[] indexes)
        {
            FileInfo fi = new FileInfo(m_documentTermMatrix);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);
            int nRprev = -1;
            bool isStarted = false;
            int nCurrentRow = -1;
            CRow[] crow = new CRow[indexes.Length];
            for (int i = 0; i < crow.Length; ++i)
            {
                crow[i] = new CRow();
            }
 
            using (FileStream stream = new FileStream(m_documentTermMatrix, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nR = reader.ReadInt32();
                        int nC = reader.ReadInt32();
                        short sum = reader.ReadInt16();

                        if (isStarted)
                        {
                            if (nR == nRprev)
                            {
                                crow[nCurrentRow].m_pos.Add(nC);
                                crow[nCurrentRow].m_sum.Add(sum);
                            }
                        }

                        if (nRprev != nR)
                        {
                            bool isFound = false;
                            foreach (int index in indexes)
                            {
                                if (nR == index)
                                {
                                    isFound = true;
                                    break;
                                }
                            }
 
                            nRprev = nR;
                            isStarted = false;

                            if (isFound)
                            {
                                ++nCurrentRow;
                                crow[nCurrentRow].m_files.Add(nR);
                                crow[nCurrentRow].m_pos.Add(nC);
                                crow[nCurrentRow].m_sum.Add(sum);
                                isStarted = true;
                            }
                        }
                    }
                }
            }
            foreach (CRow r in crow)
            {
                ComputeNorm(r);
            }
            return crow;
        }

        private CRow GetOrthogonal(CRow[] listRows)
        {
            double minCosine = 1.1 * (double)(m_trainingSet.Count);
            int nWhichRow = -1;
            for (int i = 0; i < listRows.Length; ++i)
            {
                double totalCosine = 0.0;
                for (int j = 0; j < m_trainingSet.Count; ++j)
                {
                    totalCosine += getCosine(listRows[i], m_trainingSet[j]);
                }
                if (totalCosine < minCosine)
                {
                    minCosine = totalCosine;
                    nWhichRow = i;
                }
            }
            return listRows[nWhichRow];
        }

        private void ComputeNorm(CRow row)
        {
            double norm = 0.0;
            foreach (int x in row.m_sum)
            {
                norm += (double)(x * x);
            }
            norm = Math.Sqrt(norm);
            row.norm = norm;
        }

        public void MakeTrainingSample(int number)
        {
            List<int> nCategory = new List<int>(0);
            const int nStartingCategory = 0;

            int len = m_crows[nStartingCategory].m_files.Count;
            Random rnd = new Random();
            int nWhich = rnd.Next(len);
            nWhich = m_crows[nStartingCategory].m_files[nWhich];

            int[] initial = new int[1];
            initial[0] = nWhich;
            CRow[] listR = MakeListOfRows(initial);
            m_trainingSet.Add(listR[0]);
            nCategory.Add(nStartingCategory);
 
            int nCurrentRow = 1;
            while (true)
            {
                CRow[] listRows = MakeListOfRows(m_crows[nCurrentRow].m_files.ToArray());
                CRow cr = GetOrthogonal(listRows);

                m_trainingSet.Add(cr);
                nCategory.Add(nCurrentRow);

                listRows = null;

                if (m_trainingSet.Count >= number) break;
 
                ++nCurrentRow;
                if (nCurrentRow >= m_crows.Count) nCurrentRow = 0; 
            }

            int totalCats = 0;
            foreach (int x in nCategory)
            {
                if (x > totalCats) totalCats = x;
            }
            ++totalCats;

            List<int>[] catFile = new List<int>[totalCats];
            for (int i = 0; i < totalCats; ++i)
            {
                catFile[i] = new List<int>();
            }

            int cnt = 0;
            foreach (int n in nCategory)
            {
                catFile[n].Add(m_trainingSet[cnt].m_files[0]);
                ++cnt;
            }

            Console.WriteLine("The suggested training sample.");
            Console.WriteLine("It is subject to user inspection.");
            Console.WriteLine("Should not be used as is.\n");
            for (int i = 0; i < totalCats; ++i)
            {
                Console.Write("Category {0}, files: ", i);
                foreach (int x in catFile[i])
                {
                    Console.Write(" {0} ", x);
                }
                Console.WriteLine("\n--------");
            }
        }
    }
}
