﻿/*
   Copyright 2012, Andrew Polar

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace splitandmerge
{
    class SAM
    {
        private string m_databaseFolder;
        private string m_processedFilesList;
        private string m_documentTermMatrix;
        private int m_nClusters;
        private List<Cluster> m_clusterList = new List<Cluster>();
        private List<int> m_labels = new List<int>();
        private Random m_rand = null;
        private int m_cnt = -1;
        private int m_nVectorSize = -1;
        private bool m_usePO;
        private int m_nFiles = -1;
        private HAC m_hac = null;

        public SAM(string databaseFolder, string processedFilesList, string documentTermMatrix, int nClusters, bool usePO)
        {
            m_databaseFolder = databaseFolder;
            m_processedFilesList = processedFilesList;
            m_documentTermMatrix = documentTermMatrix;
            m_nClusters = nClusters;
            m_usePO = usePO;
            m_rand = new Random();
            m_hac = new HAC(m_documentTermMatrix);
        }

        public void MakeRandomLabels()
        {
            m_nFiles = GetNumberOfFiles();
            m_labels.Clear();
            for (int i = 0; i < m_nFiles; ++i)
            {
                m_labels.Add(GetRandomLabel());
            }
        }

        private int GetRandomLabel()
        {
            if (m_cnt < m_nClusters - 1)
            {
                ++m_cnt;
                return m_cnt;
            }
            else
            {
                return m_rand.Next(m_nClusters);
            }
        }

        private int GetNumberOfFiles()
        {
            FileInfo fi = new FileInfo(m_documentTermMatrix);
            long nBytes = fi.Length;
            FileStream infile = new FileStream(m_documentTermMatrix, FileMode.Open, FileAccess.Read);
            long position = nBytes - 10;
            infile.Seek(position, SeekOrigin.Begin);
            BinaryReader reader = new BinaryReader(infile);
            int nR = reader.ReadInt32();
            reader.Close();
            infile.Close();
            return nR + 1;
        }

        private void CountNorm(ref Row row)
        {
            double norm = 0.0;
            foreach (int sum in row.m_sum)
            {
                norm += sum * sum;
            }
            norm = Math.Sqrt(norm);
            row.norm = norm;
        }

        public void MakeClusters()
        {
            for (int i = 0; i < m_clusterList.Count; ++i)
            {
                m_clusterList[i].ClearAll();
            }
            m_clusterList.Clear();

            for (int i = 0; i < m_nClusters; ++i)
            {
                m_clusterList.Add(new Cluster());              
            }

            FileInfo fi = new FileInfo(m_documentTermMatrix);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);
            int nRprev = 0;
            Row row = new Row();

            using (FileStream stream = new FileStream(m_documentTermMatrix, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nR = reader.ReadInt32();
                        int nC = reader.ReadInt32();
                        short sum = reader.ReadInt16();

                        if (m_nVectorSize < nC) m_nVectorSize = nC;

                        row.m_pos.Add(nC);
                        row.m_sum.Add(sum);

                        //Console.WriteLine("{0} {1} {2}", nR, nC, sum);

                        if (nRprev != nR)
                        {
                            CountNorm(ref row);
                            m_clusterList[m_labels[nRprev]].AddRow(row, nRprev);

                            nRprev = nR;
                            row = new Row();
                        }
                    }
                }
            }
            ++m_nVectorSize;

            CountNorm(ref row);
            m_clusterList[m_labels[nRprev]].AddRow(row, nRprev);
        }

        public void MakeSplitAndMergeStep()
        {
            for (int i = 0; i < m_clusterList.Count; ++i)
            {
                m_clusterList[i].Split(m_nVectorSize, m_usePO);
            }
        }

        public void ExecuteHAC()
        {
            m_hac.PopulateCRows(m_clusterList);
            m_hac.ExecuteHAC(m_nClusters);
            m_labels = m_hac.GetNewLabels();           
        }

        //Debugging function
        public void CountAllElements() 
        {
            int nTotal = 0;
            foreach (Cluster cluster in m_clusterList)
            {
                nTotal += cluster.CountAllElements();
            }
            Console.WriteLine("Number of elements {0}", nTotal);
        }

        public double ComputeF_measure(int[] nExpected, int nCategories)
        {
            return m_hac.ComputeF_measure(nExpected, nCategories);
        }

        public void ShowClusteredFiles()
        {
            m_hac.ShowClusteredFiles();
        }

        public void MakeTrainingSample(int number)
        {
            m_hac.MakeTrainingSample(number);
        }
    }
}
