﻿/*
   Copyright 2012, Andrew Polar

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using helper;

namespace tripleclustering
{
    class RFPlusPO
    {
        public RFPlusPO()
        {
        }

        private class m_node
        {
            public m_node leftNode = null;
            public m_node rightNode = null;
            public double[] leftClassifier = null;
            public double[] rightClassifier = null;
            public bool[] leftSample = null;
            public bool[] rightSample = null;
            public int category = -1;
        }

        m_node[] m_entryNodes = null;
        List<int> m_category = new List<int>(0);

        private double[,] m_trainingMatrix = null;
        private int[] m_categoryForTrainingData = null;
        private int m_tRows = -1;
        private int m_tCols = -1;

        private int getNumberOfCols(string dataFile, ref int nNonZeros)
        {
            FileInfo fi = new FileInfo(dataFile);
            long nBytes = fi.Length;
            nNonZeros = (int)(nBytes / 10);
            int nCols = -1;
            using (FileStream stream = new FileStream(dataFile, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nR = reader.ReadInt32();
                        int nC = reader.ReadInt32();
                        short sum = reader.ReadInt16();
                        if (nC > nCols) nCols = nC;
                    }
                    reader.Close();
                }
                stream.Close();
            }
            ++nCols;
            return nCols;
        }

        public bool buildTrainingVectors(string dataFile, string listFiles)
        {
            int[] nTabooList = Helper.getTabooList(listFiles);
            m_tRows = 0;
            foreach (int x in nTabooList)
            {
                if (x >= 0) ++m_tRows;
            }
            int nNonZeros = 0;
            m_tCols = getNumberOfCols(dataFile, ref nNonZeros);

            m_trainingMatrix = new double[m_tRows, m_tCols];
            m_categoryForTrainingData = new int[m_tRows];

            for (int i = 0; i < m_tRows; ++i)
            {
                for (int j = 0; j < m_tCols; ++j)
                {
                    m_trainingMatrix[i, j] = 0.0;
                }
            }

            int counter = 0; 
            int[] inverseTaboo = new int[nTabooList.Length];
            for (int i = 0; i < nTabooList.Length; ++i)
            {
                if (nTabooList[i] >= 0)
                {
                    inverseTaboo[i] = counter;
                    m_categoryForTrainingData[counter] = nTabooList[i];
                    ++counter;
                }
                else inverseTaboo[i] = -1;
            }

            //Populate matrix
            using (FileStream stream = new FileStream(dataFile, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nR = reader.ReadInt32();
                        int nC = reader.ReadInt32();
                        short sum = reader.ReadInt16();

                        int row = inverseTaboo[nR];
                        if (row >= 0)
                        {
                            m_trainingMatrix[row, nC] += (double)(sum);
                        }

                    }
                    reader.Close();
                }
                stream.Close();
            }
            return true;
        }

        private bool[] makeStartingSample()
        {
            bool[] startingSample = new bool[m_tRows];
            Random rnd = new Random();
            for (int i = 0; i < m_tRows; ++i)
            {
                if (rnd.Next() % 1000 < 618) startingSample[i] = true;
                else startingSample[i] = false;
            }
            return startingSample;
        }

        private void makeClassifiers(ref double[] left, ref double[] right, bool[] sample, ref bool[] leftSample, ref bool[] rightSample)
        {
            int size = 0;
            for (int i=0; i<sample.Length; ++i) 
            {
                if (sample[i]) ++size;
            }
            int[] pos = new int[size];
            int cnt = 0;
            for (int i = 0; i < sample.Length; ++i)
            {
                if (sample[i])
                {
                    pos[cnt++] = i;
                }
            }

            double[,] matrix = new double[size, size];
            for (int i = 0; i < size-1; ++i)
            {
                for (int j = i+1; j < size; ++j)
                {
                    matrix[i,j] = 0.0;
                    for (int k = 0; k < m_tCols; ++k)
                    {
                        matrix[i, j] += m_trainingMatrix[pos[i], k] * m_trainingMatrix[pos[j], k];
                    }
                }
            }

            int L = 0;
            int R = 1;
            double minValue = matrix[L, R];
            for (int i = 0; i < size - 1; ++i)
            {
                for (int j = i + 1; j < size; ++j)
                {
                    if (minValue > matrix[i, j])
                    {
                        minValue = matrix[i, j];
                        L = i;
                        R = j;
                    }
                }
            }

            List<int> leftPositions = new List<int>();
            List<int> rightPositions = new List<int>();
            leftPositions.Add(pos[L]);
            rightPositions.Add(pos[R]);

            for (int i = 0; i < size; ++i)
            {
                if (i != L && i != R)
                {
                    double sumL = 0.0;
                    for (int k = 0; k < m_tCols; ++k)
                    {
                        sumL += m_trainingMatrix[pos[L], k] * m_trainingMatrix[pos[i], k];
                    }

                    double sumR = 0.0;
                    for (int k = 0; k < m_tCols; ++k)
                    {
                        sumR += m_trainingMatrix[pos[R], k] * m_trainingMatrix[pos[i], k];
                    }

                    if (sumR > sumL)
                    {
                        rightPositions.Add(pos[i]);
                    }
                    else
                    {
                        leftPositions.Add(pos[i]);
                    }
                }
            }

            for (int i = 0; i < leftSample.Length; ++i)
            {
                leftSample[i] = false;
                rightSample[i] = false;
            }
            foreach (int x in leftPositions)
            {
                leftSample[x] = true;
            }
            foreach (int x in rightPositions)
            {
                rightSample[x] = true;
            }
            for (int k = 0; k < m_tCols; ++k)
            {
                left[k] = 0.0;
                right[k] = 0.0;
            }
            foreach (int x in leftPositions)
            {
                for (int k = 0; k < m_tCols; ++k)
                {
                    left[k] += m_trainingMatrix[x, k];
                }
            }
            foreach (int x in rightPositions)
            {
                for (int k = 0; k < m_tCols; ++k)
                {
                    right[k] += m_trainingMatrix[x, k];
                }
            }
        }

        private void Orthogonalize(ref double[] left, ref double[] right)
        {
            double norm = 0.0;
            int size = left.Length;
            for (int i = 0; i < size; ++i)
            {
                norm += left[i] * left[i];
                norm += right[i] * right[i];
            }
            norm = Math.Sqrt(norm);
            for (int i = 0; i < size; ++i)
            {
                left[i] /= norm;
                right[i] /= norm;
            }
            double[,] tmp = new double[2, 2];
            double[] tmpLeft = new double[m_tCols];
            double[] tmpRight = new double[m_tCols];

            int counter = 0;
            while (true)
            {
                tmp[0, 0] = 0.0;
                tmp[0, 1] = 0.0;
                tmp[1, 0] = 0.0;
                tmp[1, 1] = 0.0;
                for (int i = 0; i < size; ++i)
                {
                    tmp[0, 0] += left[i] * left[i];
                    tmp[0, 1] += left[i] * right[i];
                    tmp[1, 1] += right[i] * right[i];
                }
                tmp[1, 0] = tmp[1, 0];

                for (int i = 0; i < size; ++i)
                {
                    tmpLeft[i] = left[i] * tmp[0, 0] + right[i] * tmp[0, 1];
                    tmpRight[i] = left[i] * tmp[1, 0] + right[i] * tmp[1, 1];
                }

                for (int i = 0; i < size; ++i)
                {
                    left[i] = 1.5 * left[i] - 0.5 * tmpLeft[i];
                    right[i] = 1.5 * right[i] - 0.5 * tmpRight[i];
                }

                double side = 2.0 * tmp[0, 1];
                double trace = tmp[0, 0] + tmp[1, 1];
                if (side / trace < 0.00001) break;
                if (++counter >= 100) break;
            }

            /*for (int i = 0; i < size; ++i)
            {
                if ((left[i] > 0 && right[i] > 0) || (left[i] < 0 && right[i] < 0))
                {
                    if (Math.Abs(left[i] - right[i]) / Math.Abs(left[i] + right[i]) < 0.2)
                    {
                        left[i] /= 2.0;
                        right[i] /= 2.0;
                    }
                }
            }*/
        }

        private m_node nodeFactory(bool[] sample)
        {
            m_node node = new m_node();
            Dictionary<int, int> dictionary = new Dictionary<int, int>();
            int nCat = -1;
            for (int i = 0; i < sample.Length; ++i)
            {
                if (sample[i])
                {
                    if (!dictionary.ContainsKey(m_categoryForTrainingData[i])) 
                    {
                        nCat = m_categoryForTrainingData[i];
                        dictionary.Add(m_categoryForTrainingData[i], i);
                    }
                }
            }
            if (dictionary.Count == 1)
            {
                node.category = nCat;
                return node;
            }

            node.leftClassifier = new double[m_tCols];
            node.rightClassifier = new double[m_tCols];
            node.leftSample = new bool[sample.Length];
            node.rightSample = new bool[sample.Length];

            makeClassifiers(ref node.leftClassifier, ref node.rightClassifier, sample, ref node.leftSample, ref node.rightSample);
            Orthogonalize(ref node.leftClassifier, ref node.rightClassifier);

            return node;
        }

        private void makeLeftAndRightNodes(m_node node)
        {
            node.leftNode = nodeFactory(node.leftSample);
            node.rightNode = nodeFactory(node.rightSample);
            if (node.leftNode.category < 0)
            {
                makeLeftAndRightNodes(node.leftNode);
            }
            if (node.rightNode.category < 0)
            {
                makeLeftAndRightNodes(node.rightNode);
            }
        }

        public void buidForest(int N)
        {
            m_entryNodes = new m_node[N];
            for (int k = 0; k < N; ++k)
            {
                bool[] sample = makeStartingSample();
                m_entryNodes[k] = nodeFactory(sample);
                if (m_entryNodes[k].category < 0)
                {
                    makeLeftAndRightNodes(m_entryNodes[k]);
                }
                Console.Write("{0}  \r", k + 1);
            }
        }

        private int getCategoryFromTheTree(double[] vector, m_node node)
        {
            if (node.category >= 0) return node.category;
            double fLeft = 0.0;
            double fRight = 0.0;
            for (int k = 0; k < vector.Length; ++k)
            {
                fLeft += node.leftClassifier[k] * vector[k];
                fRight += node.rightClassifier[k] * vector[k];
            }
 
            if (fLeft > fRight) return getCategoryFromTheTree(vector, node.leftNode);
            else return getCategoryFromTheTree(vector, node.rightNode);
        }

        private int getCategory(double[] vector, int nCategories)
        {
            int[] catCollection = new int[nCategories];
            for (int i = 0; i < catCollection.Length; ++i) catCollection[i] = 0;
            for (int k = 0; k < m_entryNodes.Length; ++k)
            {
                int cat = getCategoryFromTheTree(vector, m_entryNodes[k]);
                ++catCollection[cat];
            }

            int maxPos = 0;
            int max = catCollection[maxPos];
            for (int k = 0; k < catCollection.Length; ++k)
            {
                if (catCollection[k] > max)
                {
                    max = catCollection[k];
                    maxPos = k;
                }
            }
            return maxPos;
        }

        public void processVectors(string dataFile, int nCategories)
        {
            m_category.Clear();
            FileInfo fi = new FileInfo(dataFile);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);
            int nCurrentRow = 0;
            double[] vector = new double[m_tCols];
            for (int m = 0; m < m_tCols; ++m) vector[m] = 0.0;
            using (FileStream stream = new FileStream(dataFile, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nR = reader.ReadInt32();
                        int nC = reader.ReadInt32();
                        short sum = reader.ReadInt16();

                        if (nR == nCurrentRow)
                        {
                            vector[nC] = sum;
                        }
                        else
                        {
                            m_category.Add(getCategory(vector, nCategories));
                            for (int m = 0; m < m_tCols; ++m) vector[m] = 0.0;
                            nCurrentRow = nR;
                        }
                    }
                    reader.Close();
                }
                stream.Close();
            }
            m_category.Add(getCategory(vector, nCategories));
        }

        public double computeF_measure(int[] nExpected, int nCategories)
        {
            return Helper.Compute_F_Measure(nExpected, m_category.ToArray(),  true, "RF");
        }

        public int[] getResult()
        {
            return m_category.ToArray();
        }
    }
}
